/*
 *	USELESS - Driver Project	
 *  This component facilitates the actual driver keyboard LED control.
 *
 *  Copyright (c)2003 Brad Basler, All Rights Reserved.
 */


#include <ntddk.h>
#include "driver.h"





NTSTATUS DriverEntry( 
    IN PDRIVER_OBJECT  DriverObject, 
    IN PUNICODE_STRING  RegistryPath 
)
{

	NTSTATUS ntStatus;

	UNICODE_STRING UniSZ_DeviceName;
	UNICODE_STRING UniSZ_DeviceLink;

	PDEVICE_OBJECT pDeviceObject;

	PDEVICE_EXTENSION pDeviceExtension;

	//Initialize Device Name
	RtlInitUnicodeString(&UniSZ_DeviceName,L"\\Device\\Useless");
	
	//Create device driver
	ntStatus = IoCreateDevice(
					DriverObject,
					sizeof(DEVICE_EXTENSION),				//FIX - Device Extension Size
					&UniSZ_DeviceName,
					USELESS_BASE_FILE_DEVICE_TYPE,
					0,
					FALSE,
					&pDeviceObject
			   );

	if (NT_SUCCESS(ntStatus))
	{

		//Setup driver/OS interfaces.

		pDeviceExtension = (PDEVICE_EXTENSION)pDeviceObject->DeviceExtension;

		RtlZeroMemory(pDeviceExtension,sizeof(DEVICE_EXTENSION));

		//Store the pointer of our object in the extension
		pDeviceExtension->pDeviceObject	= pDeviceObject;

		DriverObject->MajorFunction[IRP_MJ_CLOSE]			=		Useless_Dispatch;
		DriverObject->MajorFunction[IRP_MJ_CREATE]			=		Useless_Dispatch;
		DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]	=		Useless_Dispatch;

		DriverObject->DriverUnload							=		Useless_Unload;

		//Setup our symbolic link
		RtlInitUnicodeString(&UniSZ_DeviceLink,L"\\DosDevices\\Useless_LED_Controller");

		//Link
		ntStatus = IoCreateSymbolicLink(
						&UniSZ_DeviceLink,
						&UniSZ_DeviceName
				   );

		if (NT_SUCCESS(ntStatus))
		{
			pDeviceExtension->SuccessMask |= SUCCESS_SYMBOLIC_LINK;

			//Success - Load Complete

		}
		else
		{
			DbgPrint("Useless: Error! - Symbolic Link Creation Failed!!\n");

			//Delete the loaded driver object
			IoDeleteDevice(pDeviceObject);
		}

	}
	else DbgPrint("Useless: Error! - Driver Device could not be created\n");
	
	//Return Success/Failure.
	return ntStatus;

}


NTSTATUS Useless_Dispatch(PDEVICE_OBJECT DeviceObject,PIRP Irp)
{

	PIO_STACK_LOCATION IrpStack;
	ULONG              dwInputBufferLength;
	ULONG              dwOutputBufferLength;
	ULONG              dwIoControlCode;
	PVOID              pvIOBuffer;
	NTSTATUS           ntStatus;

	Irp->IoStatus.Status		= STATUS_SUCCESS;
	Irp->IoStatus.Information	= 0;

	IrpStack = IoGetCurrentIrpStackLocation(Irp);

	pvIOBuffer				= Irp->AssociatedIrp.SystemBuffer;
	dwInputBufferLength		= IrpStack->Parameters.DeviceIoControl.InputBufferLength;
	dwOutputBufferLength	= IrpStack->Parameters.DeviceIoControl.OutputBufferLength;

	switch (IrpStack->MajorFunction)
	{
		case IRP_MJ_DEVICE_CONTROL:
			{

				dwIoControlCode = IrpStack->Parameters.DeviceIoControl.IoControlCode;

				switch (dwIoControlCode)
				{

					case IOCTL_SET_LEDS:

						if ((dwInputBufferLength != 1) || (dwOutputBufferLength))
							Irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
						else
						{

							UCHAR NewLeds = (UCHAR)*(PUCHAR)pvIOBuffer;

							SetLEDS(NewLeds);

						}
						break;
				}

			}
			break;

		default:
			break;

	}

	ntStatus = Irp->IoStatus.Status;

	IoCompleteRequest(Irp,IO_NO_INCREMENT);

	return ntStatus;
}

void Useless_Unload(PDRIVER_OBJECT pDriverObject)
{
	if (pDriverObject && pDriverObject->DeviceObject && pDriverObject->DeviceObject->DeviceExtension)
	{

		PDEVICE_EXTENSION pDeviceExtension = (PDEVICE_EXTENSION)pDriverObject->DeviceObject->DeviceExtension;

		if (pDeviceExtension->SuccessMask & SUCCESS_SYMBOLIC_LINK)
		{

			UNICODE_STRING UniSZ_SymbolicLink;

			RtlInitUnicodeString(&UniSZ_SymbolicLink,L"\\DosDevices\\Useless_LED_Controller");

			IoDeleteSymbolicLink(&UniSZ_SymbolicLink);

		}

		IoDeleteDevice(pDriverObject->DeviceObject);

	}

}

////////////////////////////////////////////////////////////////////
// Code Borrowed from Richard Hoglund's basic hardware example
//
////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////

// commands
#define READ_CONTROLLER		0x20
#define WRITE_CONTROLLER	0x60

// command bytes
#define SET_LEDS			0xED
#define KEY_RESET			0xFF

// responses from keyboard
#define KEY_ACK				0xFA	// ack
#define KEY_AGAIN			0xFE	// send again

// 8042 ports
// when you read from port 64, this is called STATUS_BYTE
// when you write to port 64, this is called COMMAND_BYTE
// read and write on port 64 is called DATA_BYTE 
PUCHAR KEYBOARD_PORT_60 = (PUCHAR)0x60;
PUCHAR KEYBOARD_PORT_64 = (PUCHAR)0x64;

// status register bits
#define IBUFFER_FULL		0x02
#define OBUFFER_FULL		0x01

// flags for keyboard LEDS
#define SCROLL_LOCK_BIT		(0x01 << 0)
#define NUMLOCK_BIT			(0x01 << 1)
#define CAPS_LOCK_BIT		(0x01 << 2)

ULONG WaitForKeyboard()
{
	int i = 100;	// number of times to loop
	UCHAR mychar;
	
	do
	{
		mychar = READ_PORT_UCHAR( KEYBOARD_PORT_64 );

		KeStallExecutionProcessor(666);

		if(!(mychar & IBUFFER_FULL)) break;	// if the flag is clear, we go ahead
	}
	while (i--);

	if(i) return TRUE;
	return FALSE;
}


// call WaitForKeyboard before calling this function
void DrainOutputBuffer()
{
	int i = 100;	// number of times to loop
	UCHAR c;
	do
	{
		c = READ_PORT_UCHAR(KEYBOARD_PORT_64);
		
		KeStallExecutionProcessor(666);

		if(!(c & OBUFFER_FULL)) break;	// if the flag is clear, we go ahead
	
		// gobble up the byte in the output buffer
		c = READ_PORT_UCHAR(KEYBOARD_PORT_60);
		
	}
	while (i--);
}

// write a byte to the data port at 0x60
ULONG SendKeyboardCommand( IN UCHAR theCommand )
{
	if(TRUE == WaitForKeyboard())
	{
		DrainOutputBuffer();

		WRITE_PORT_UCHAR( KEYBOARD_PORT_60, theCommand );
	}
	else return FALSE;
	
	// TODO: wait for ACK or RESEND from keyboard	
	
	return TRUE;
}

void SetLEDS( UCHAR theLEDS )
{
	// setup for setting LEDS
	if (SendKeyboardCommand( 0xED ) == TRUE)
		SendKeyboardCommand( theLEDS );
}